
/*

DPMIREP: DPMI report program, V1.0
Copyright (c) 1992, UK, John M.Howells - All rights reserved

This program can be built with the compiler from either the
Borland C++ 3.1 or Microsoft C/C++ 7.0 packages. It can
probably be built with others, but that has not been checked.

Borland C++3.1:		bcc -ms -O1 dpmirep.c
Microsoft C/C++ 7.0:	cl -AS -Os -Gs dpmirep.c

This program *must* be compiled in small mode, as it relies on the
data segment not having to be modified while the program is running.
Also, using the usual exit() function causes an exception error, so
a macro is declared to make sure you remember this!

This program only supports DPMI 0.90, as so far I have not come
across a DPMI 1.00 host. The information should be valid under
a DPMI 1.00 host, but a different function, 0x050B, is preferred
for getting memory information from a 1.00 host. Under DPMI 1.00
the program could also get more information with function 0x0401.

DISTRIBUTION
============

This program is provided as freeware. You are allowed to distribute it
in any way, as long as no charge is made beyond any out of pocket expenses
involved in such distribution (up to a maximum of 3). Any changes to this
source must retain the original copyright in this source and in the output,
and must include an attribution to the modifying author in both places, in
addition to clear comments in the source indicating the modified code.
Any modified source must also contain this notice and must be distributed
free of charge, on the same basis a this original. Note in particular that
the program is *not* public domain. Any distribution must include the
source for the program as well as the executable.

						John M.Howells
						27th August 1992
						CompuServe 100031,353
*/


#include	<stdio.h>
#include	<dos.h>
#include	<string.h>


#define	UNS		unsigned
#define	PARA_SIZE	16


/* interrupts */
#define	INT_DOS			0x21		/* DOS services */
#define	INT_MPX			0x2F		/* Multiplex */
#define	INT_DPMI		0x31		/* DPMI host services */

/* values for DOS interrupt */
#define	DOS_GETMEM		0x48		/* Get block of DOS memory */
#define	DOS_EXIT		0x4C		/* Terminate Program */

/* values for multiplex interrupt */
#define	TSW_CHK			0x4680		/* to check for a Task Swtch */
#define	WIN_ENH_CHK		0x1600		/* to check Enh Windows */
#define	GET_DPMI_ENT		0x1687		/* to get Prot mode entry pt */

/* values for DPMI interrupts */
#define	DPMI_GET_VERS		0x0400		/* Get DPMI version */
#define	DPMI_GET_MEMINFO_0_90	0x0500		/* Get DPMI 0.90 Mem Info */
#define	DPMI_GET_PAG_SIZ	0x0604		/* Get DPMI Page size */

/* bits in DPMI flags word */
#define	DPMI_FLG_32_BIT		0x0001		/* 32-bit mode ? */
#define	DPMI_FLG_INT_MODE	0x0002		/* reflected interrupts ? */
#define	DPMI_FLG_VM_SUPP	0x0004		/* VM supported ? */


#define	dim( xx ) 	( ( sizeof( xx ) / sizeof( xx[0] ) ) - 1 )


/* force a linker 'undefined symbol' error if exit() is used.
   We could just map exit(xx) to dos_exit(xx), but We want to
   remind anyone doing mods that there is more to exit() than
   dos_exit() when operating under DPMI. */
#define	exit( xx )	do_not_use_the_usual_exit( xx )


/* We cannot use exit() so we do out own */
void	dos_exit( UNS char err_num )
{ 
	/* because exit() is not called there is no automatic
	   flush of the output buffers, so it is done here */
	fflush( stdout );

	/* this works in real or protected mode */
	_asm	mov	al, err_num
	_asm	mov	ah, DOS_EXIT
	_asm	int	INT_DOS
}


/* Report if Windows in Enhanced mode */
void	win_enh( void )
{
	UNS char maj_vers;
	UNS char min_vers;

	_asm	mov	ax, WIN_ENH_CHK
	_asm	int	INT_MPX
	_asm	mov	maj_vers, al
	_asm	mov	min_vers, ah

	if( maj_vers &&
	   (! (maj_vers == 0x80 || maj_vers == 1 || maj_vers == 0xFF)) )
	{
		/* not only do we know Windows is running in Enhanced mode, but
		   the undocumented check has also told us the version number */
		printf( "The DPMI host is Windows %d.%2.2d in Enhanced mode\n\n",
				maj_vers,
				min_vers );
	}
}


/* function to report fatal error and quit */
void	quit( char *s )
{
	puts( s );
	dos_exit( 1 );
}


/* check if a task switcher - or Standard mode Windows - running */
void	tsw_run( void )
{
	/* if you don't do this check then using this program under Windows
	   Standard mode or DOS 5's DOSSHELL if the Task Swithcher is enabled
	   results in a reboot of the machine if QuarterDeck's QDPMI 1.0 is
	   enabled!

	   And it ain't just this program, because it also happens if you use
	   Borland's C++ 3.1 or MS' C/C++ 7.0 compilers. Funny things also
	   happen if you use 386MAX as the host for these compilers and have
	   DOSSHELL's Task Switcher enabled, so it's not be QDPMI's problem.
	   If DOSSHELL's Task Switcher is *not* enabed it's all okay.

	   Note: that this must be the *first* check, because if you get to
	   the DPMI host check it's too late, as that blows up. */

	_asm	mov	ax, TSW_CHK	/* Windows 3.x Real or Standard or .. */
	_asm	int	INT_MPX		/* .. DOSSHELL task switcher check */
	_asm	or	ax,ax		/* check return value */
	_asm	jnz	no_tsw		/* if non zero task swtch not running */

	quit( "A Task Switcher is running, or Windows but not in Enhanced mode" );

no_tsw:
	;				/* MS' C/C++ 7.0 wants this,
					   Borland's C++ 3.1 doesn't care */
}


/* Get the DPMI Switch Entry Point address so the program can get
   from real to protected mode, check if DPMI is available, and
   switch to protected mode by calling the switch entry point. */
int	dpmi_start( void )
{
	UNS host_data_size;
	UNS dpmi_flags;
	void (far *dpmi_sw)();			/* set up to entry ptr */

	dpmi_flags = 0;				/* 16-bit prot-mode prog */

	_asm	mov	ax, GET_DPMI_ENT	/* is DPMI present */
	_asm	int	INT_MPX
	_asm	or	ax, ax
	_asm	jz	dpmi			/* AX == 0 - DPMI */

	quit( "There is no DPMI host available" );

dpmi:
	_asm	mov	word ptr dpmi_sw, di	/* save entry ptr */
	_asm	mov	word ptr dpmi_sw + 2, es

	_asm	mov	host_data_size, si	/* host private data size */

	_asm	mov	bx, si			/* allocate paragraphs */
	_asm	mov	ah, DOS_GETMEM
	_asm	int	INT_DOS
	_asm	jnc	mem_ok

	quit( "Can't allocate memory for host private data" );

mem_ok:	
	_asm	mov	es, ax			/* ES = data segment */
	_asm	mov	ax, dpmi_flags

	_asm	call	dword ptr dpmi_sw	/* go to protected mode */

	_asm	jnc	dpmi_ok

	quit( "Failed to switch to protected (DPMI) mode" );

dpmi_ok:
	return( host_data_size );
}


#ifdef	__BORLANDC__			/* Borland C++ 3.1's pragma to .. */
#pragma	warn	-rvl			/* ignore "no return value" warning */
#endif

#ifdef	_MSC_VER			/* MS C/C++ 7.0's pragma to .. */
#pragma	warning( disable:4035 )		/* ignore "no return value" warning */
#endif

main()
{
	UNS dpmi_vers;
	UNS dpmi_flgs;
	UNS proc_type;
	UNS int_nums;
	UNS ms_pgsiz, ls_pgsiz;
	UNS long page_size;
	UNS host_paras;
	static	struct {
		UNS long	largest_block;
		UNS long	maximum_unlock_pages;
		UNS long	maximum_locked_pages;
		UNS long	linear_address_pages;
		UNS long	tot_unlocked_pages;
		UNS long	tot_free_pages;
		UNS long	tot_physical_pages;
		UNS long	free_linear_space;
		UNS long	size_of_paging_file;
		UNS long	filler[3];
	} mem_info;

	static	char	*form_1 = "%-50.50s %10lu\n";	/* format for one number */
	static	char	*form_2 = "%-50.50s %10lu (%lu)\n";	/* for two numbers */
	static	char	*proc_names[] = { "Unrecognized (type 0)",
					  "Unrecognized (type 1)",
					  "80286",
					  "80386",
					  "80486" };
	static	char	proc_str[100];

	printf( "\n"
		"DPMIREP: DPMI report program V1.0\n"
		"Copyright (c) 1992, UK, John M.Howells - All rights reserved\n"
		"\n" );	

	tsw_run();
	host_paras = dpmi_start();
	    
	/* if we get here we are in protected mode */
	
	win_enh();

	_asm	mov	ax, DPMI_GET_VERS	/* Get Version */
	_asm	int	INT_DPMI
	_asm	mov	dpmi_vers, ax		/* and save results */
	_asm	mov	dpmi_flgs, bx
	_asm	mov	proc_type, cx
	_asm	mov	int_nums, dx

	if( dpmi_flgs & DPMI_FLG_32_BIT )
	{
		_asm	mov	ax, DPMI_GET_PAG_SIZ
		_asm	int	INT_DPMI
		_asm	mov	ms_pgsiz, bx
		_asm	mov	ls_pgsiz, cx
		_asm	jnc	okay_page

		page_size = 0;
		goto	 page_bad;

	okay_page:
		page_size = ( ms_pgsiz * 65536L ) + ls_pgsiz;

	page_bad:
		;			/* MS' C/C++ 7.0 wants this,
					   Borland's C++ 3.1 doesn't care */
	}
	else
	{
		page_size = 0;
	}

	_asm	mov	ax, ds
	_asm	mov	es, ax
	_asm	mov	di, offset mem_info
	_asm	mov	ax, DPMI_GET_MEMINFO_0_90
	_asm	int	INT_DPMI

	_asm	jnc	okay_mem

	quit( "failed to get mem info\n" );

okay_mem:
	if( ( proc_type & 0xFF ) > dim( proc_names ) )
	{
		sprintf( proc_str, "Unrecognized (type %d)", proc_type & 0xFF );
	}
	else
	{
		strcpy( proc_str, proc_names[ ( proc_type & 0xFF ) ] );
	}
	printf( "DPMI Version is %d.%2.2d on an %s processor\n",
		( dpmi_vers >> 8 ) & 0xFF, dpmi_vers & 0xFF, proc_str );
	printf( "DPMI host implementation is %s\n",
		( dpmi_flgs & DPMI_FLG_32_BIT ) ?
				"32-bit (80386 or better)" : "16-bit" );
	printf( "CPU returned to %s mode for reflected interrupts\n",
		( dpmi_flgs & DPMI_FLG_INT_MODE ) ?
				"real" : "Virtual 86" );
	printf( "Virtual memory is %s\n",
		( dpmi_flgs & DPMI_FLG_VM_SUPP ) == 0 ?
			"not supported" :
			( mem_info.size_of_paging_file == (UNS long)-1 ||
			  mem_info.size_of_paging_file == (UNS long) 0 ) ?
				"supported - but not enabled" :
			 	"supported" );
	printf( "Current Virtual PIC base interrupt: Master = 0x%2.2X, Slave = 0x%2.2X\n",
		( int_nums >> 8 ), (int_nums & 0xFF ) );
	printf( "DPMI host requires %lu bytes (%u paras) for private data\n",
		(UNS long)host_paras * PARA_SIZE,
		host_paras );

	printf( "\n" );

	if( page_size )
	{
		printf( form_1, "SIze of one page in bytes", page_size );
	}
	else
	{
		if( dpmi_flgs & DPMI_FLG_32_BIT )
		{
			printf( "Failed to get page size\n\n" );
		}
	}

	if( mem_info.largest_block != (UNS long)-1 )
	{
		printf( form_1, "Largest available free block in bytes",
			mem_info.largest_block );
	}
	if( mem_info.maximum_unlock_pages != (UNS long)-1 )
	{
		printf( form_1, "Maximum unlocked page allocation in pages",
			mem_info.maximum_unlock_pages );
	}
	if( mem_info.maximum_locked_pages != (UNS long)-1 )
	{
		printf( form_1, "Maximum locked page allocation in pages",
			mem_info.maximum_locked_pages );
	}
	if( mem_info.linear_address_pages != (UNS long)-1 )
	{
		printf( form_2, "Linear address space size in pages (bytes)",
			mem_info.linear_address_pages,
			mem_info.linear_address_pages * page_size );
	}
	if( mem_info.tot_unlocked_pages != (UNS long)-1 )
	{
		printf( form_1, "Total number of unlocked pages",
			mem_info.tot_unlocked_pages );
	}
	if( mem_info.tot_free_pages != (UNS long)-1 )
	{
		printf( form_1, "Total number of free pages",
			mem_info.tot_free_pages );
	}
	if( mem_info.tot_physical_pages != (UNS long)-1 )
	{
		printf( form_1, "Total number of physical pages",
			mem_info.tot_physical_pages );
	}
	if( mem_info.free_linear_space != (UNS long)-1 )
	{
		printf( form_2, "Free linear address space in pages (bytes)",
			mem_info.free_linear_space,
			mem_info.free_linear_space * page_size );
	}
	if( mem_info.size_of_paging_file != (UNS long)-1 )
	{
		printf( form_2, "Size of paging file/partition in pages (bytes)",
			mem_info.size_of_paging_file,
			mem_info.size_of_paging_file * page_size );
	}

	dos_exit( 0 );
}

